# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import json
import pandas as pd
import uuid

from dash_extensions.enrich import html, dash_table

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep


class ProjectInformationTableAIO(html.Div):

    class ids:
        datatable = lambda aio_id: {
            'component': 'ProjectInformationTableAIO',
            'subcomponent': 'datatable',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, monitoring_step: MonitoringStep, datatable_props: dict = None, interval_props: dict = None, aio_id: str = None):
        """ProjectInformationTableAIO is an All-in-One component that is composed
        of a parent `html.Div` with a `dcc.Interval` and a `dash_table.DataTable` as children.

        - `monitoring_step` - The StepModel object of the monitoring step.
        - `datatable_props` - A dictionary of properties passed into the dash_table.DataTable component.
        - `interval_props` - A dictionary of properties passed into the dcc.Interval component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        data = self.get_data(monitoring_step)

        datatable_props = {
            "data": data.to_dict('records'),
            "columns": [{"name": i, "id": i, "type": "text"} for i in data.columns],
            "fixed_rows": {"headers": True},
            "style_header": {
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": "15px",
                "fontWeight": "bold",
                "border": "none",
                "display": "none",
            },
            "style_cell": {
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": "15px",
                "border": "none"
            },
            "style_as_list_view": True,
            "style_table": {"border": "none"}, # Hide table borders
        }

        super().__init__([
            dbc.Card(
                [
                    dbc.CardBody(
                        [
                            html.H4("Project information", className="card-title"),
                            dash_table.DataTable(id=self.ids.datatable(aio_id), **datatable_props),
                        ]
                    ),
                ],
                color="warning",
                outline=True,
            )
        ])

    def get_data(self, monitoring_step: MonitoringStep) -> pd.DataFrame:

        project_status_info = json.loads(monitoring_step.project_status_info_file.read_text())

        project_summary_data = {
            "column_a": [
                "State",
                "Id",
                "Name",
                "Machine",
                "Location",
                "Project directory",
                "Owner",
                "Registered",
                "Lock info",
            ],
            "column_b": [
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
            ],
        }

        if project_status_info:
            project_summary_data["column_b"] = [
                project_status_info["projects"][0]["state"],
                project_status_info["projects"][0]["project_id"],
                project_status_info["projects"][0]["name"],
                project_status_info["projects"][0]["machine"],
                project_status_info["projects"][0]["location"],
                project_status_info["projects"][0]["working_dir"],
                project_status_info["projects"][0]["user"],
                "",
                "",
            ]

        return pd.DataFrame(project_summary_data)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Command Line Interface for PyAnsys Templates."""

import os

import click

from ansys.templates import AVAILABLE_TEMPLATES_AND_DESCRIPTION, __version__
from ansys.templates.paths import (
    PYTHON_TEMPLATES_OSL_SOLUTION_PATH,
    PYTHON_TEMPLATES_SOLUTION_PATH,
    TEMPLATE_PATH_FINDER,
)
from ansys.templates.utils import bake_template, load_inputs_from_configuration_file


def create_project(template, no_input=False, extra_context={}):
    """Create Python project based on a given template.

    Parameters
    ----------
    template : str
        Name of the template to be used as basis for the project

    """
    bake_template(TEMPLATE_PATH_FINDER[template], os.getcwd(), overwrite_if_exists=True, no_input=no_input, extra_context=extra_context)


@click.group()
def main():
    """Ansys tool for creating new Ansys projects."""
    pass


@main.command()
def list():
    """List all available templates names."""
    print("Available templates in ansys-templates are:\n")
    for template_name, description in AVAILABLE_TEMPLATES_AND_DESCRIPTION.items():
        print(f"{template_name.replace('_', '-')}: {description}")


@main.command()
def version():
    """Display current version."""
    print(f"ansys-templates {__version__}")


@main.group()
def new():
    """Create a new project from desired template."""
    pass


@new.command()
def doc_project():
    """Create a documentation project using Sphinx."""
    create_project("doc-project")


@new.command()
def pybasic():
    """Create a basic Python Package."""
    create_project("pybasic")


@new.command()
def pyansys():
    """Create a PyAnsys Python Package project."""
    create_project("pyansys")


@new.command()
def pyansys_advanced():
    """Create an advanced PyAnsys Python Package project."""
    create_project("pyansys-advanced")


@new.command()
def pyansys_openapi_client():
    """Create an OpenAPI Client Package project."""
    create_project("pyansys-openapi-client")


@new.command()
def pyace():
    """Create a Python project for any method developers."""
    create_project("pyace")


@new.command()
def pyace_fast():
    """Create a FastAPI project initialized for any developer."""
    create_project("pyace-fast")


@new.command()
def pyace_flask():
    """Create a Flask project initialized for any developer."""
    create_project("pyace-flask")


@new.command()
def pyace_grpc():
    """Create gRPC project initialized for any developer."""
    create_project("pyace-grpc")

@new.command()
@click.option('-f',  '--from', '_from' , help='From existing optiSLang application archive (OWA).',
              type=click.Choice(['owa'], case_sensitive=False))
@click.option('-s', '--solution-name', type=str, help="Name of the solution in the definition.")
@click.option('-d', '--solution-display-name', type=str, help="Name of the solution in the user interface.")
@click.option("-u", "--with-dash-ui", is_flag=True, flag_value="1", help="With Dash UI")
@click.option('-a', '--application-archive', type=click.Path(), help="Path to the optiSLang application archive.")
def solution(_from, solution_name, solution_display_name, with_dash_ui, application_archive):
    """[Ansys Internal Use Only] Create a solution based on SAF."""
    if _from == 'owa':
        template = "osl-solution"
        extra_context = load_inputs_from_configuration_file(PYTHON_TEMPLATES_OSL_SOLUTION_PATH)
        no_input = True if application_archive else False
        if application_archive:
            extra_context["optiSLang_application_archive"] = application_archive
    else:
        template = "solution"
        extra_context = load_inputs_from_configuration_file(PYTHON_TEMPLATES_SOLUTION_PATH)
        no_input = True if solution_name or solution_display_name or with_dash_ui else False
        if solution_name:
            extra_context["solution_name"] = solution_name
        if solution_display_name:
            extra_context["solution_display_name"] = solution_display_name
        if with_dash_ui:
            extra_context["with_dash_ui"] = "yes"

    create_project(template, no_input=no_input, extra_context=extra_context)



{%- if cookiecutter.copyright != "None" -%}
# Copyright (c) {% now "utc", '%Y' %}, {{ cookiecutter.copyright }}. Unauthorised use, distribution or duplication is prohibited
{% endif %}

"""
{{ cookiecutter.project_name }}.

{{ cookiecutter.library_name }}
"""
import argparse

from flask import Flask
from flask_cors import CORS
from flask_swagger_ui import get_swaggerui_blueprint

from blueprints.health import blueprint as health_endpoint
from blueprints.version import blueprint as version_endpoint
from observability.logger import Logger

SWAGGER_URL = "/api/docs"
API_URL = "/static/swagger.json"


SWAGGER_UI_BLUEPRINT = get_swaggerui_blueprint(
    SWAGGER_URL,
    API_URL,
    config={"app_name": "{{ cookiecutter.project_name }} {{ cookiecutter.library_name }} Api"},
)

logger = Logger.init("{{ cookiecutter.__project_name_slug }}")


def create_app():
    """Initialize the core application."""
    app = Flask(__name__)
    CORS(app, resources=r"/api/*")
    app.config["CORS_HEADERS"] = "Content-Type"
    app.register_blueprint(version_endpoint)
    app.register_blueprint(health_endpoint)
    app.register_blueprint(SWAGGER_UI_BLUEPRINT, url_prefix=SWAGGER_URL)

    return app


def serve(app, address, port, middleware=None):
    """Serve the application."""
    if middleware is not None:
        middleware(app)

    logger.info("{{ cookiecutter.project_name }} {{ cookiecutter.library_name }} Server starting...")
    app.run(host=address, port=port)


if __name__ == "__main__":
    app = create_app()
    logger.info("server.py main : parsing arguments")
    parser = argparse.ArgumentParser()
    parser.add_argument("--address", help="Set server address", default="0.0.0.0")
    parser.add_argument("-p", "--port", type=int, help="Set server port", default=5000)
    args = parser.parse_args()
    serve(app=app, address=args.address, port=args.port)



{%- if cookiecutter.copyright != "None" -%}
# Copyright (c) {% now "utc", '%Y' %}, {{ cookiecutter.copyright }}. Unauthorised use, distribution or duplication is prohibited
{% endif %}
"""
{{ cookiecutter.project_name }}.

{{ cookiecutter.library_name }}
"""

from flask import Blueprint, jsonify

from observability.logger import Logger

blueprint = Blueprint("health_check", __name__, url_prefix="/api/health")

logger = Logger.init("{{ cookiecutter.__project_name_slug }}")


@blueprint.route("/")
def health_check():
    """Check health status."""
    logger.info("Health check")
    return jsonify({"status": "ok"})



{%- if cookiecutter.copyright != "None" -%}
# Copyright (c) {% now "utc", '%Y' %}, {{ cookiecutter.copyright }}. Unauthorised use, distribution or duplication is prohibited
{% endif %}
"""
{{ cookiecutter.project_name }}.

{{ cookiecutter.library_name }}
"""

from flask import Blueprint, jsonify

from _version import __version__

blueprint = Blueprint("api_version", __name__, url_prefix="/api/version")


@blueprint.route("/")
def get_version():
    """Get current version."""
    return jsonify({"version": __version__})



import shutil
import os
from pathlib import Path
from ansys.templates.utils import keep_files
import zipfile


DESIRED_STRUCTURE = [
    ".github/workflows/ci.yml",
    ".vscode/launch.json",
    "doc/source/_static/ansys-solutions-logo-black-background.png",
    "doc/source/_static/README.md",
    "doc/source/_templates/README.md",
    "doc/source/conf.py",
    "doc/source/index.rst",
    "doc/styles/Vocab/ANSYS/accept.txt",
    "doc/styles/Vocab/ANSYS/reject.txt",
    "doc/styles/.gitignore",
    "doc/.vale.ini",
    "doc/make.bat",
    "doc/Makefile",
    "examples/README.md",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/model/assets/README.md",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/pim_configurations/osl_wrapper.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/pim_configurations/pim_osl_configuration.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/solution/method_assets/README.md",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/solution/definition.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/solution/problem_setup_step.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/solution/monitoring_step.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/solution/optislang_manager.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/assets/css/style.css",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/assets/images/README.md",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/assets/logos/ansys-solutions-horizontal-logo.png",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/assets/scripts/README.md",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/actor_information_table.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/actor_logs_table.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/actor_statistics_table.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/button_group.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/node_control.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/design_table.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/logs_table.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/project_information_table.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/service_table.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/components/system_files.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/pages/monitoring_page.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/pages/page.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/pages/problem_setup_page.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/utils/alerts.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/views/design_table_view.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/views/project_summary_view.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/views/result_files_view.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/views/scenery_view.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/views/status_overview_view.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/views/summary_view.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/views/visualization_view.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/ui/app.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/utilities/common_functions.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/__init__.py",
    f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/main.py",
    "telemetry/grafana/compose.yaml",
    "telemetry/grafana/_deploy/compose/collector/otel-collector-config.yaml",
    "telemetry/grafana/_deploy/compose/grafana/grafana-dashboard.json",
    "telemetry/grafana/_deploy/compose/grafana/grafana-dashboards.yml",
    "telemetry/grafana/_deploy/compose/grafana/grafana-datasources.yml",
    "telemetry/grafana/_deploy/compose/grafana/grafana.ini",
    "telemetry/grafana/_deploy/compose/loki/loki.yml",
    "telemetry/grafana/_deploy/compose/prometheus/prometheus.yml",
    "telemetry/tracelens/compose.yaml",
    "tests/common_test_files/README.md",
    "tests/integration/test_integration_dummy.py",
    "tests/unit/test_unit_dummy.py",
    "tests/conftest.py",
    ".codespell.exclude",
    ".codespell.ignore",
    ".env",
    ".flake8",
    ".gitignore",
    ".pre-commit-config.yaml",
    "AUTHORS",
    "CHANGELOG.md",
    "CODE_OF_CONDUCT.md",
    "CODEOWNERS",
    "CONTRIBUTING.md",
    "CONTRIBUTORS.md",
    "LICENSE.rst",
    "poetry.lock",
    "pyproject.toml",
    "README.rst",
    "setup_environment.py",
    "tox.ini"
]
"""A list holding all desired files to be included in the project."""

LOGIC_ASSETS_DIRECTORY = Path(f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/logic/assets").absolute()
METHOD_ASSETS_DIRECTORY = Path(f"src/ansys/solutions/{{ cookiecutter.__solution_name_slug }}/solution/method_assets/Problem_Setup").absolute()

def unzip_archive(archive_path: Path, extract_path: Path) -> None:
    """Unzip an archive."""

    with zipfile.ZipFile(archive_path, "r") as zip_ref:
        zip_ref.extractall(extract_path)


def copy_file_to_assets_folder(file_path: str, destination: str) -> None:
    """Copy a file if it exists."""

    if not os.path.isabs(file_path):
        working_directory = os.path.dirname(os.getcwd())
        file_path = os.path.join(working_directory, file_path)

    if os.path.exists(file_path):
        shutil.copy(file_path, destination)
    else:
        print(f"Unable to find {file_path}.")


def collect_files_with_extension(directory, extension):
    """
    Collects all files with the specified extension from a directory.
    Args:
        directory (str): The directory path.
        extension (str): The desired file extension.
    Returns:
        list: A list of file names with the specified extension.
    """
    files_with_extension = [file for file in os.listdir(directory) if file.endswith(extension)]
    return files_with_extension


def main():
    """Entry point of the script."""

    keep_files(DESIRED_STRUCTURE)

    if len("{{ cookiecutter.__optiSLang_application_archive }}".replace(" ", "")):
        archive_path = Path("{{ cookiecutter.__optiSLang_application_archive }}".strip())

        if not archive_path.is_absolute():
            raise Exception("Relative path not allowed, please provide the absolute path of the owa archive.")
        if not archive_path.exists():
            raise Exception(f"File not found: {archive_path}")

        unzip_archive(archive_path, METHOD_ASSETS_DIRECTORY / "{{ cookiecutter.__optiSLang_application_archive_stem }}")

        for file in ["metadata.json", "doc.md"]:
            copy_file_to_assets_folder(
                str(METHOD_ASSETS_DIRECTORY / "{{ cookiecutter.__optiSLang_application_archive_stem }}" / file),
                str(METHOD_ASSETS_DIRECTORY / file)
            )
        for extension in [".json", ".opf"]:
            candidates = collect_files_with_extension(str(METHOD_ASSETS_DIRECTORY / "{{ cookiecutter.__optiSLang_application_archive_stem }}" / "custom_data"), extension)
            if len(candidates) == 0:
                raise Exception("The optiSLang application archive contains no project file (opf).")
            elif len(candidates) > 1:
                raise Exception("The optiSLang application archive contains multiple project files (opf).")
            else:
                candidate = "{{ cookiecutter.__optiSLang_application_archive_stem }}" + extension
                if extension == ".json":
                    destination_folder = METHOD_ASSETS_DIRECTORY
                if extension == ".opf":
                    destination_folder = LOGIC_ASSETS_DIRECTORY
                copy_file_to_assets_folder(
                    str(METHOD_ASSETS_DIRECTORY / "{{ cookiecutter.__optiSLang_application_archive_stem }}" / "custom_data" / candidates[0]),
                    str(destination_folder / candidate)
                )
        shutil.rmtree(str(METHOD_ASSETS_DIRECTORY / "{{ cookiecutter.__optiSLang_application_archive_stem }}"))


if __name__ == "__main__":
    main()



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import json
import pandas as pd
import uuid

from dash_extensions.enrich import html, dash_table

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import sorted_nicely


class DesignTableAIO(html.Div):

    class ids:
        datatable = lambda aio_id: {
            'component': 'DesignTableAIO',
            'subcomponent': 'datatable',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, monitoring_step: MonitoringStep, aio_id: str = None):
        """DesignTableAIO is an All-in-One component that is composed
        of a parent `html.Div` with a `dcc.Interval` and a `dash_table.DataTable` as children.

        - `monitoring_step` - The StepModel object of the monitoring step.
        - `datatable_props` - A dictionary of properties passed into the dash_table.DataTable component.
        - `interval_props` - A dictionary of properties passed into the dcc.Interval component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        data = self.get_data(monitoring_step)

        datatable_props = {
            "data": data.to_dict('records'),
            "columns": [{"name": i, "id": i, "type": "text"} for i in data.columns],
            "fixed_rows": {"headers": True},
            "style_header": {
                "textAlign": "center",
                "font_family": "Roboto",
                "font_size": "15px",
                "fontWeight": "bold",
            },
            "style_cell": {
                "textAlign": "center",
                "font_family": "Roboto",
                "font_size": "15px",
                "whiteSpace": "normal",
            },
            "style_data_conditional": [
                {
                    "if": {"column_id": "Status", "filter_query": '{Status} eq "Succeeded"'},
                    "backgroundColor": "rgb(223, 240, 208)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Status", "filter_query": '{Status} eq "Not succeeded"'},
                    "backgroundColor": "rgb(254, 221, 215)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "True"'},
                    "backgroundColor": "rgb(223, 240, 208)",
                    "color": "rgb(0, 0, 0)",
                },
                {
                    "if": {"column_id": "Feasible", "filter_query": '{Feasible} eq "False"'},
                    "backgroundColor": "rgb(254, 221, 215)",
                    "color": "rgb(0, 0, 0)",
                },
            ],
            "style_as_list_view": True,
        }

        super().__init__([
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(
                        dash_table.DataTable(id=self.ids.datatable(aio_id), **datatable_props),
                        width=12
                    ),
                ]
            )
        ])

    def sort_designs(self, designs: list) -> list:

        sorted_designs = sorted_nicely(designs)
        return [designs.index(design) for design in sorted_designs]

    def get_data(self, monitoring_step: MonitoringStep) -> pd.DataFrame:

        actors_info = json.loads(monitoring_step.actors_info_file.read_text())
        actors_status_info = json.loads(monitoring_step.actors_status_info_file.read_text())

        if monitoring_step.selected_actor_from_treeview in actors_info.keys():
            actor_info = actors_info[monitoring_step.selected_actor_from_treeview]
        else:
            actor_info = {}

        if monitoring_step.selected_actor_from_treeview in actors_status_info.keys():
            actor_status_info = actors_status_info[monitoring_step.selected_actor_from_treeview][0]
        else:
            actor_status_info = {}

        design_table_data = {
            "Design": [],
            "Feasible": [],
            "Status": [],
            "Pareto": [],
        }

        field_types = ["parameter", "constraint", "objective", "response"]

        if actor_info:
            if actor_info["kind"] == "system":
                for field_type in field_types:
                    field_name_index = field_type + "_names"
                    for field_name in actor_status_info["designs"][field_name_index]:
                        design_table_data[field_name] = []
                design_numbers = []
                for design_values in reversed(actor_status_info["designs"]["values"]):
                    design_numbers.append(design_values["hid"].split(".")[-1])
                    design_table_data["Design"].append(design_values["hid"])
                    for field_type in field_types:
                        field_name_index = field_type + "_names"
                        field_value_index = field_type + "_values"
                        field_names = actor_status_info["designs"][field_name_index]
                        field_values = design_values[field_value_index]
                        if len(field_names) != len(field_values):
                            field_values = [None for i in range(len(field_names))]
                        for field_name, field_value in zip(field_names, field_values):
                            if isinstance(field_value, dict):
                                if field_value["type"] == "xy_data":
                                    field_value = f"[1:%s]" % (field_value["num_entries"])
                            design_table_data[field_name].append(field_value)
                for design_status in reversed(actor_status_info["design_status"]):
                    design_table_data["Feasible"].append(design_status["feasible"])
                    design_table_data["Status"].append(design_status["status"])
                    design_table_data["Pareto"].append(design_status["pareto_design"])

                design_table_data = pd.DataFrame(design_table_data)

                # OptiSLang returns the designs in an unsorted way.
                # The following method organizes the rows in ascending order.
                design_table_data = design_table_data.loc[self.sort_designs(design_table_data["Design"].to_list())]

                for col, dtype in design_table_data.dtypes.items():
                    if dtype == "bool":
                        design_table_data[col] = design_table_data[col].astype("str")

                design_table_data = design_table_data.round(6)

        return pd.DataFrame(design_table_data)



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Logs table"""

import pandas as pd

from dash_extensions.enrich import dash_table
from typing import Union


class LogsTable:
    """Logs table."""

    def __init__(self, data: Union[dict, list]) -> None:
        """Constructor."""

        self._data = data
        self.font_size: str = "15px"

    def _reshape_data(self):

        if isinstance(self._data, dict):
            pass
        elif not any(isinstance(i, list) for i in self._data):
            # This is a non-nested list.
            self._data = {"column_a": self._data}
        else:
            raise ValueError("Unprocessable data. Expect a non-nested list or a dictionary.")

    def render(self):
        """Generate table."""

        self._reshape_data()

        data = pd.DataFrame(self._data)

        return dash_table.DataTable(
            data=data.to_dict("records"),
            columns=[{"name": i, "id": i, "type": "text"} for i in data.columns],
            fixed_rows={"headers": True},
            style_header={
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": self.font_size,
                "fontWeight": "bold",
                "display": "none",
            },
            style_cell={
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": self.font_size,
            },
            style_data_conditional=[
                {
                    "if": {"row_index": "odd"},
                    "backgroundColor": "rgb(242, 242, 242)",
                },
                {
                    "if": {"row_index": "even"},
                    "backgroundColor": "rgb(255, 255, 255)",
                },
            ],
            style_as_list_view=True,
        )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import json
import pandas as pd
import plotly.express as px
import uuid

from dash_extensions.enrich import html, dash_table, dcc

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep


class ActorInformationTableAIO(html.Div):

    class ids:
        datatable = lambda aio_id: {
            'component': 'ActorInformationTableAIO',
            'subcomponent': 'datatable',
            'aio_id': aio_id
        }
        pie_chart = lambda aio_id: {
            'component': 'ActorStatisticsTableAIO',
            'subcomponent': 'pie_chart',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, monitoring_step: MonitoringStep, aio_id: str = None):
        """ActorInformationTableAIO is an All-in-One component that is composed
        of a parent `html.Div` with a `dcc.Interval` and a `dash_table.DataTable` as children.

        - `monitoring_step` - The StepModel object of the monitoring step.
        - `datatable_props` - A dictionary of properties passed into the dash_table.DataTable component.
        - `interval_props` - A dictionary of properties passed into the dcc.Interval component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        table_data, pie_chart_data = self.get_data(monitoring_step)

        datatable_props = {
            "data": table_data.to_dict('records'),
            "columns": [{"name": i, "id": i, "type": "text"} for i in table_data.columns],
            "fixed_rows": {"headers": True},
            "style_header": {
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": "15px",
                "fontWeight": "bold",
                "border": "none",
                "display": "none",
            },
            "style_cell": {"textAlign": "left", "font_family": "Roboto", "font_size": "15px", "border": "none"},
            "style_as_list_view": True,
            "style_table": {"border": "none"},  # Hide table borders
        }

        if pie_chart_data:
            pie_chart_props = {
                "figure": px.pie(
                    pie_chart_data,
                    values='Count',
                    names='Category',
                    color='Category',
                    color_discrete_map={
                        'Success': 'rgba(173, 191, 11, 1)',
                        'Failure': 'rgba(248, 105, 80, 1)',
                        'Pending': 'rgba(251, 201, 93, 1)'
                    },
                ).update_traces(textinfo='none', hole=0.8, showlegend=True),
                "style": {
                    "width": "300px",
                    "height": "300px",
                    "display": "inline-block",
                },
            }

        super().__init__([
            dbc.Card(
                [
                    dbc.Row(
                        [
                            dbc.Col(
                                [
                                    html.H4("Actor information", className="card-title"),
                                    dash_table.DataTable(id=self.ids.datatable(aio_id), **datatable_props),
                                ],
                                width=8
                            ),
                            dbc.Col(
                                [
                                    dcc.Graph(id=self.ids.pie_chart(aio_id), **pie_chart_props) if pie_chart_data else None,
                                ],
                                width=2
                            ),
                        ]
                    )
                ],
                color="warning",
                outline=True,
            )
        ])

    def get_data(self, monitoring_step: MonitoringStep) -> pd.DataFrame:

        actors_info = json.loads(monitoring_step.actors_info_file.read_text())
        actors_status_info = json.loads(monitoring_step.actors_status_info_file.read_text())

        if monitoring_step.selected_actor_from_treeview in actors_info.keys():
            actor_info = actors_info[monitoring_step.selected_actor_from_treeview]
        else:
            actor_info = {}

        if monitoring_step.selected_actor_from_treeview in actors_status_info.keys():
            actor_status_info = actors_status_info[monitoring_step.selected_actor_from_treeview][0]
        else:
            actor_status_info = {}

        actor_information_data = {
            "column_a": ["Working directory", "Processing state", "Execution duration"],
            "column_b": [None, None, None],
        }

        pie_chart_data = None

        if actor_info and actor_status_info:
            actor_information_data["column_b"] = [
                actor_status_info["working dir"],
                actor_status_info["state"],
                "-",
            ]
            if actor_info["kind"] == "system":
                succeeded_designs = actor_status_info["succeeded_designs"]
                failed_designs = actor_status_info["failed_designs"]
                pending_designs = actor_status_info["pending_designs"]
                total_designs = actor_status_info["total_designs"]
                processed_designs = total_designs - pending_designs
                status = int(processed_designs / total_designs * 100)
                actor_information_data["column_a"].extend(["Processed", "Status", "Succeeded", "Not succeeded"])
                actor_information_data["column_b"].extend(
                    [f"{processed_designs} / {total_designs}", f"{status}%", succeeded_designs, failed_designs]
                )

                pie_chart_data = {'Category': ['Success', 'Failure', 'Pending'], 'Count': [0, 0, 0]}
                for index, state in enumerate(["succeeded_designs", "failed_designs", "pending_designs"]):
                    pie_chart_data["Count"][index] = actor_status_info[state]

        return pd.DataFrame(actor_information_data), pie_chart_data



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import uuid

from dash_extensions.enrich import html, dcc

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import (
    remove_key_from_dictionaries,
    sort_dict_by_ordered_keys,
)


class SystemFilesAIO(html.Div):

    class ids:
        button_omdb = lambda aio_id: {
            'component': 'SystemFilesAIO',
            'subcomponent': 'button_omdb',
            'aio_id': aio_id
        }
        button_csv = lambda aio_id: {
            'component': 'SystemFilesAIO',
            'subcomponent': 'button_csv',
            'aio_id': aio_id
        }
        download_omdb = lambda aio_id: {
            'component': 'SystemFilesAIO',
            'subcomponent': 'download_omdb',
            'aio_id': aio_id
        }
        download_csv = lambda aio_id: {
            'component': 'SystemFilesAIO',
            'subcomponent': 'download_csv',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, monitoring_step: MonitoringStep, aio_id: str = None):
        """SystemFilesAIO is an All-in-One component that is composed
        of a parent `html.Div` with a `dcc.Interval` and a `dash_table.DataTable` as children.

        - `monitoring_step` - The StepModel object of the monitoring step.
        - `datatable_props` - A dictionary of properties passed into the dash_table.DataTable component.
        - `interval_props` - A dictionary of properties passed into the dcc.Interval component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        button_props = {}
        download_props = {}

        super().__init__([
            dbc.Card(
                [
                    dbc.CardBody(
                        [
                            html.H4("System Files", className="card-title"),
                            dbc.Stack(
                                [
                                    html.Div(
                                        [
                                            dbc.Button(
                                                children="Download OMDB",
                                                id=self.ids.button_omdb(aio_id),
                                                **button_props
                                            ),
                                            dcc.Download(id=self.ids.download_omdb(aio_id), **download_props),
                                        ]
                                    ),
                                    html.Div(
                                        [
                                            dbc.Button(
                                                children="Download CSV",
                                                id=self.ids.button_csv(aio_id),
                                                **button_props
                                            ),
                                            dcc.Download(id=self.ids.download_csv(aio_id), **download_props),
                                        ]
                                    ),
                                ],
                                direction="horizontal",
                                gap=3,
                            )
                        ]
                    )
                ],
                color="warning",
                outline=True,
            )
        ])



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the first step."""


from ansys.saf.glow.client.dashclient import DashClient, callback
from dash_extensions.enrich import Input, Output, State, dcc, html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{cookiecutter.__solution_definition_name}}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.first_step import FirstStep
from ansys.solutions.dash_components.table import InputRow,OutputRow


def layout(step: FirstStep):
    """Layout of the first step UI."""
    return html.Div(
        [
            dcc.Markdown("""#### First step""", className="display-3"),
            dcc.Markdown("""###### Compute the sum of two numbers.""", className="display-3"),
            html.Hr(className="my-2"),
            html.Br(),
            InputRow(
                "number",
                "first-arg",
                "First Argument",
                row_description='Enter a value.',
                label_width=2,
                value_width=4,
                unit_width=1,
                description_width=4,
                font_size="16px"
            ).get(),
            InputRow(
                "number",
                "second-arg",
                "Second Argument",
                row_description='Enter a value.',
                label_width=2,
                value_width=4,
                unit_width=1,
                description_width=4,
                font_size="16px"
            ).get(),
            InputRow(
                "button",
                "calculate",
                "Calculate",
                label_width=2,
                value_width=4,
                unit_width=1,
                description_width=4,
                class_name="button",
                font_size="16px"
            ).get(),
            html.Br(),
            OutputRow(
                "number",
                "result",
                "Result",
                row_default_value=step.result,
                label_width=2,
                value_width=4,
                unit_width=1,
                description_width=4,
                class_name="button",
                font_size="16px"
            ).get(),
            dcc.Loading(
                type="circle",
                fullscreen=True,
                color="#ffb71b",
                style={
                    "background-color": "rgba(55, 58, 54, 0.1)",
                },
                children=html.Div(id="wait_completion"),
            ),
        ]
    )


# This is an example callback which stores entered data and executes a method.
# This is the only place where user entered data is persisted in this Dash app.
# Notice that the project is referenced by a call to get_project and
# the url is a State argument to the callback
@callback(
    Output("result", "value"),
    Output("wait_completion", "children"),
    Input("calculate", "n_clicks"),
    State("first-arg", "value"),
    State("second-arg", "value"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def calculate(n_clicks, first_arg, second_arg, pathname):
    """Callback function to trigger the computation."""
    project = DashClient[{{cookiecutter.__solution_definition_name}}].get_project(pathname)
    step = project.steps.first_step
    step.first_arg = first_arg
    step.second_arg = second_arg
    step.calculate()
    return step.result, True



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the first step."""


from ansys.saf.glow.solution import StepModel, StepSpec, transaction


class FirstStep(StepModel):
    """Step definition of the first step."""

    first_arg: float = 0
    second_arg: float = 0
    result: float = 0

    @transaction(self=StepSpec(upload=["result"], download=["first_arg", "second_arg"]))
    def calculate(self) -> None:
        """Method to compute the sum of two numbers."""
        self.result = self.first_arg + self.second_arg



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import json
import dash_bootstrap_components as dbc
import uuid

from dash_extensions.enrich import Output, Input, State, html, dcc, callback, MATCH, callback_context, ALL
from dash.exceptions import PreventUpdate

from ansys.saf.glow.client.dashclient import DashClient
from ansys.saf.glow.core.method_status import MethodStatus

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.button_group import ButtonGroup

class NodeControlAIO(html.Div):

    class ids:
        button_group = lambda aio_id: {
            'component': 'NodeControlAIO',
            'subcomponent': 'btn_group',
            'aio_id': aio_id
        }
        alert = lambda aio_id: {
            'component': 'NodeControlAIO',
            'subcomponent': 'alert',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, button_group, alert_props, aio_id: str = None):
        """NodeControlAIO is an All-in-One component that is composed
        of a parent `html.Div` with a `dbc.Stack` and a `dbc.Alert` as children.

        - `button_group` - The container of buttons used to build the project commands.
        - `alert_props` - The dictionary of properties for the alert component.
        - `aio_id` - The All-in-One component ID used to generate the components dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        super().__init__([
            dbc.Card(
                [
                    dbc.CardBody(
                        [
                            html.H4("Project commands", className="card-title"),
                            dbc.Row(
                                [
                                    dbc.Col(
                                        dbc.Stack(
                                            id=self.ids.button_group(aio_id),
                                            children=button_group,
                                            direction="horizontal",
                                            gap=1,
                                        ),
                                        width="auto",
                                    ),
                                    dbc.Col(
                                        children=[
                                            dbc.Alert(
                                                id=self.ids.alert(aio_id),
                                                dismissable=True,
                                                fade= True,
                                                style= {"width": "600px", "height": "40px", "display": "flex", "justify-content": "left", "align-items": "center"},
                                                **alert_props,
                                            ),
                                        ],
                                        width="auto",
                                    ),
                                    html.Div(id="action-requested", style={"display": "none"}, children=False),
                                ],
                            ),

                        ]
                    ),
                ],
                color="warning",
                outline=True,
            )
        ])

    @callback(
        Output(ids.alert(MATCH), "children"),
        Output(ids.alert(MATCH), "color"),
        Output(ids.alert(MATCH), "is_open"),
        Output(ids.button_group(MATCH), "children"),
        Input("action-requested", "children"),
        State("url", "pathname"),
        prevent_initial_call=True
    )
    def run_command(action_requested, pathname):
        """This performs actions such as stopping, restarting, and shutting down on an instance of optiSlang when action is requested. A message with the status of the method execution and states of each button is updated on completion of the method."""
        project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
        monitoring_step = project.steps.monitoring_step
        monitoring_step.selected_actor_from_command = monitoring_step.selected_actor_from_treeview
        ctx = callback_context
        if (ctx.triggered and action_requested):
            alert_status = {"alert-message": "", "alert-color": ""}
            try:
                monitoring_step.control_node_state()
                command_execution_state = monitoring_step.get_method_state("control_node_state").status
                if command_execution_state == MethodStatus.Completed:
                    alert_status["alert-message"] =  f"{monitoring_step.selected_command.replace('_', ' ').title()} command completed successfully.",
                    alert_status["alert-color"] = "success"
            except Exception as e:
                alert_status["alert-message"] = f"{monitoring_step.selected_command.replace('_', ' ').title()} command failed. For details, review the logs.",
                alert_status["alert-color"] = "warning"
        else:
            raise PreventUpdate
        if monitoring_step.actor_uid:
            monitoring_step.actor_command_execution_status = alert_status
            monitoring_step.project_command_execution_status = {"alert-message": "", "alert-color": ""}
            btn_group_options = monitoring_step.actor_btn_group_options
        else:
            monitoring_step.project_command_execution_status = alert_status
            monitoring_step.actor_command_execution_status = {"alert-message": "", "alert-color": ""}
            btn_group_options = monitoring_step.project_btn_group_options
        monitoring_step.commands_locked = False
        return alert_status["alert-message"], alert_status["alert-color"], bool(alert_status["alert-message"]), ButtonGroup(options=btn_group_options, disabled=monitoring_step.commands_locked).buttons

    @callback(
        Output("action-requested", "children"),
        Output({"type": "action-button", "action": ALL}, 'disabled'),
        Input({"type": "action-button", "action": ALL}, 'n_clicks'),
        State({"type": "action-button", "action": ALL}, 'disabled'),
        State("url", "pathname"),
        prevent_initial_call=True,
    )
    def set_selected_command_and_disable_buttons(n_clicks_actions, button_states, pathname):
        """This disables all project and actor command buttons on click of a button and sends and returns a bool for action requested."""
        project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
        monitoring_step = project.steps.monitoring_step
        monitoring_step.selected_actor_from_command = monitoring_step.selected_actor_from_treeview
        ctx = callback_context
        if (
            ctx.triggered
            and not all(n_click is None for n_click in n_clicks_actions)
        ):
            triggered_button = callback_context.triggered[0]["prop_id"].split(".")[0]
            action = json.loads(triggered_button)["action"]
            monitoring_step.selected_command = action
            disable_buttons = [True] * len(button_states)
            monitoring_step.commands_locked = True
        else:
            raise PreventUpdate
        return True, disable_buttons



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the project summary step."""

import dash_bootstrap_components as dbc
import json

from dash_extensions.enrich import html, Input, Output, State, dcc

from ansys.saf.glow.client.dashclient import DashClient, callback

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.node_control import NodeControlAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.project_information_table import ProjectInformationTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.service_table import ServiceTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.button_group import ButtonGroup


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the project summary view."""
    alert_props = {
        "children":monitoring_step.project_command_execution_status["alert-message"],
        "color":monitoring_step.project_command_execution_status["alert-color"],
        "is_open": bool(monitoring_step.project_command_execution_status["alert-message"]),
    }
    button_group = ButtonGroup(options=monitoring_step.project_btn_group_options, disabled=monitoring_step.commands_locked).buttons

    # Get project data
    project_data = json.loads(problem_setup_step.project_data_file.read_text())

    return html.Div(
        [
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(html.Div(ProjectInformationTableAIO(monitoring_step), id="project_information_table"), width=8),
                    dbc.Col(html.Div(ServiceTableAIO(problem_setup_step), id="service_table"), width=4),
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(NodeControlAIO(button_group, alert_props, "project-commands"), width=12),
                ]
            ),
            dcc.Interval(
                id="project_summary_auto_update",
                interval=monitoring_step.auto_update_frequency,  # in milliseconds
                n_intervals=0,
                disabled=False if monitoring_step.auto_update_activated else True
            ),
        ]
    )


@callback(
    Output("project_summary_auto_update", "disabled"),
    Input("activate_auto_update", "on"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def activate_auto_update(on, pathname):
    """Enable/Disable auto update."""
    return not on


@callback(
    Output("project_information_table", "children"),
    Output("selected_state_dropdown", "options"),
    Output("selected_state_dropdown", "value"),
    Input("project_summary_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_view(n_intervals, pathname):
    """Update design table."""
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step
    # Get monitoring step
    monitoring_step = project.steps.monitoring_step

    return ProjectInformationTableAIO(monitoring_step)



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import json
import pandas as pd
import uuid

from dash_extensions.enrich import html, dash_table
from datetime import datetime

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import (
    remove_key_from_dictionaries,
    sort_dict_by_ordered_keys,
)


class ActorLogsTableAIO(html.Div):

    class ids:
        datatable = lambda aio_id: {
            'component': 'ActorLogsTableAIO',
            'subcomponent': 'datatable',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, monitoring_step: MonitoringStep, aio_id: str = None):
        """ActorLogsTableAIO is an All-in-One component that is composed
        of a parent `html.Div` with a `dcc.Interval` and a `dash_table.DataTable` as children.

        - `monitoring_step` - The StepModel object of the monitoring step.
        - `datatable_props` - A dictionary of properties passed into the dash_table.DataTable component.
        - `interval_props` - A dictionary of properties passed into the dcc.Interval component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        data = self.get_data(monitoring_step)

        datatable_props = {
            "data": data.to_dict('records'),
            "columns": [{"name": i, "id": i, "type": "text"} for i in data.columns],
            "fixed_rows": {"headers": True},
            "style_header": {"font_family": "Roboto", "font_size": "15px", "fontWeight": "bold"},
            "style_cell": {
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": "15px",
            },
            "style_cell_conditional": [
                {"if": {"column_id": "Time"}, "minWidth": "60px", "maxWidth": "60px", "width": "60px"},
                {
                    "if": {"column_id": "Level"},
                    "minWidth": "30px",
                    "maxWidth": "30px",
                    "width": "30px",
                    "textAlign": "center",
                },
                {"if": {"column_id": "Message"}, "minWidth": "200px", "maxWidth": "200px", "width": "200px"},
            ],
            "style_data_conditional": [
                {
                    "if": {"column_id": "Level", "filter_query": '{Level} eq "INFO"'},
                    "backgroundColor": "rgb(227, 245, 252)",
                    "color": "rgb(0, 0, 0)",
                    "textAlign": "center",
                }
            ],
            "style_as_list_view": True,
        }

        super().__init__([
            dbc.Card(
                [
                    dbc.CardBody(
                        [
                            html.H4("Actor Log", className="card-title"),
                            dash_table.DataTable(id=self.ids.datatable(aio_id), **datatable_props),
                        ]
                    ),
                ],
                color="warning",
                outline=True,
            )
        ])

    def get_data(self, monitoring_step: MonitoringStep) -> pd.DataFrame:

        actors_info = json.loads(monitoring_step.actors_info_file.read_text())

        if monitoring_step.selected_actor_from_treeview in actors_info.keys():
            actor_info = actors_info[monitoring_step.selected_actor_from_treeview]
        else:
            actor_info = {}

        has_data = False

        if actor_info:
            if "log_messages" in actor_info.keys():
                if len(actor_info["log_messages"]):
                    has_data = True
                    # Remove hid key from list dictionaries because it is useless for UI
                    # and prevent transformation to DataFrame.
                    actor_logs_data = remove_key_from_dictionaries(actor_info["log_messages"], "hid")
                    # Transform list of dictionaries into dictionary
                    actor_logs_data = {
                        key: [d[key] for d in actor_info["log_messages"]]
                        for key in actor_info["log_messages"][0]
                    }
                    # Sort keys in order
                    actor_logs_data = sort_dict_by_ordered_keys(actor_logs_data, ["time_stamp", "level", "message"])
                    # Rename keys
                    actor_logs_data["Time"] = actor_logs_data.pop("time_stamp")
                    actor_logs_data["Level"] = actor_logs_data.pop("level")
                    actor_logs_data["Message"] = actor_logs_data.pop("message")
                    # Convert timestamp
                    for index, value in enumerate(actor_logs_data["Time"]):
                        dt = datetime.strptime(value, "%Y%m%dT%H%M%S.%f")
                        actor_logs_data["Time"][index] = dt.strftime("%Y-%m-%d %H-%M-%S-%f")[:-3]

        if not has_data:
            actor_logs_data = {"Time": [], "Level": [], "Message": []}

        return pd.DataFrame(actor_logs_data)



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Initialization of the frontend layout across all the steps."""

import dash_bootstrap_components as dbc
import json
import webbrowser

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Input, Output, State, callback_context, dcc, html

from ansys.saf.glow.client.dashclient import DashClient, callback
from ansys.saf.glow.core.method_status import MethodStatus
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.logs_table import LogsTable
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.pages import monitoring_page, problem_setup_page
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import extract_dict_by_key
from ansys_web_components_dash import AwcDashTree


layout = html.Div(
    [
        dcc.Location(id="url", refresh=False), # represents the browser address bar and doesn't render anything
        html.Div(
            id="progress_bar",
            style={
                'margin-left': '35%', 'margin-top': '25%'
            },
        ),
        dcc.Interval(
            id="progress_bar_update",
            interval=1000,
            n_intervals=0,
            disabled=False
        ),
        dcc.Store(id='trigger_layout_display'),
        dcc.Store(id='trigger_treeview_display'),
        dcc.Store(id='trigger_body_display'),
        html.Div(id="page_layout")
    ]
)


@callback(
    Output("trigger_layout_display", "data"),
    Input("url", "pathname"),
)
def initialization(pathname):
    """Run methods to initialize the solution."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)

    if not project.steps.problem_setup_step.project_initialized:
        project.steps.problem_setup_step.upload_bulk_files_to_project_directory()
        project.steps.problem_setup_step.get_app_metadata()
        project.steps.problem_setup_step.get_default_placeholder_values()
        project.steps.problem_setup_step.project_initialized = True

    raise PreventUpdate


@callback(
    Output("trigger_layout_display", "data"),
    Output("progress_bar", "children"),
    Output("progress_bar", "style"),
    Output("progress_bar_update", "disabled"),
    Input("progress_bar_update", "n_intervals"),
    State("url", "pathname"),
)
def update_progress_bar(n_intervals, pathname):
    """Track status of solution initialization."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    if not problem_setup_step.project_initialized:

        completion_rate = 0
        message = None
        methods = ["upload_bulk_files_to_project_directory", "get_app_metadata", "get_default_placeholder_values"]

        for method in methods:
            status = problem_setup_step.get_method_state(method).status
            if status == MethodStatus.Completed:
                completion_rate += 1
            elif status == MethodStatus.Running:
                message = method.replace("_", " ").capitalize()
                break

        completion_rate = round(completion_rate / len(methods) * 100)

        return (
            True,
            [
                dbc.Progress(
                    value=completion_rate,
                    label=f"{completion_rate} %",
                    color="rgba(255,182,35,1)",
                    style={"width": "600px", "height": "30px"},
                ),
                dbc.Label(message),
            ],
            {'margin-left': '35%', 'margin-top': '25%'},
            False
        )

    else:
        return True, [], {"display": "none"}, True


@callback(
    Output("trigger_body_display", "data"),
    Output("page_layout", "children"),
    Input("url", "pathname"),
    Input("trigger_layout_display", "data"),
    prevent_initial_call=True,
)
def display_page_layout(pathname, trigger_layout_display):
    """Display page layout."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    if problem_setup_step.project_initialized:
        return (
            True,
            [
                dbc.Stack(
                [
                    html.Div(
                        [
                            html.Img(
                            src = r"/assets/logos/ansys-solutions-horizontal-logo.png",
                            style={'width': '80%'}
                        )
                        ],
                    ),
                    html.Div(
                        [
                            dbc.Button(
                                f"Project Name: {project.project_display_name}",
                                id = "project_name",
                                disabled = True,
                                style={
                                    "color": "rgba(0, 0, 0, 1)",
                                    "background-color": "rgba(255, 255, 255, 1)",
                                    "border-color": "rgba(0, 0, 0, 1)"
                                },
                            )
                        ],
                        className="ms-auto",
                    ),
                    html.Div(
                        [
                            dbc.Button(
                                "Back to Projects",
                                id = "return_to_portal",
                                className = "me-2",
                                n_clicks = 0,
                                href=DashClient.get_portal_ui_url(),
                                style = {"background-color": "rgba(0, 0, 0, 1)", "border-color": "rgba(0, 0, 0, 1)"},
                            ) if DashClient.get_portal_ui_url() else []
                        ],
                    ),
                ],
                direction = "horizontal",
                gap = 3,
            ),
                html.Br(),
                dbc.Row(
                    [
                        dbc.Col(
                            [
                                AwcDashTree(
                                    id="navigation_tree",
                                    multi=False,
                                    height=950,
                                    items=problem_setup_step.treeview_items,
                                    selectedItemIds=["problem_setup"]
                                ),

                            ],
                            width=2,
                            style={"background-color": "rgba(242, 242, 242, 0.6)"},  # Ansys grey
                        ),
                        dbc.Col(
                            [
                                dbc.Row(
                                    html.Div(
                                        id="body_content",
                                        style={"padding-right": "2%"}
                                    ),
                                ),
                                html.Br(),
                                dbc.Row(
                                    html.Div(
                                        id="bottom_button_group",
                                        style={"padding-right": "2%"}
                                    ),
                                ),
                            ],
                            width=10
                        ),
                    ],
                ),
            ]
        )
    else:
        raise PreventUpdate


@callback(
    Output("body_content", "children"),
    Output("bottom_button_group", "children"),
    Output("navigation_tree", "focusRequested"),
    Input("navigation_tree", "treeItemClicked"),
    Input("url", "pathname"),
    Input("trigger_body_display", "data"),
    prevent_initial_call=True,
)
def display_body_content(value, pathname, trigger_body_display):
    """Display body content."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step
    monitoring_step = project.steps.monitoring_step

    if problem_setup_step.project_initialized:
        triggered_id = callback_context.triggered[0]["prop_id"].split(".")[0]
        footer_buttons = [
            dbc.Button(
                "optiSLang logs",
                id="osl_logs_button",
                n_clicks=0,
                style={"background-color": "rgba(242, 242, 242, 0.6)", "borderColor": "rgba(242, 242, 242, 0.6)", "color": "rgba(0, 0, 0, 1)"},
                size="sm",
            ),
        ]
        if DashClient.get_portal_ui_url():
            footer_buttons.append(
                dbc.Button(
                    "Open in browser",
                    id="open_in_browser",
                    style={"background-color": "rgba(242, 242, 242, 0.6)", "borderColor": "rgba(242, 242, 242, 0.6)", "color": "rgba(0, 0, 0, 1)"},
                    n_clicks=0,
                    size="sm"
                )
            )
        if triggered_id == "url" or triggered_id == "trigger_body_display" or trigger_body_display and len(triggered_id) == 0:
            page_layout = problem_setup_page.layout(problem_setup_step)
            focusRequested = ""
        if triggered_id == "navigation_tree":
            focusRequested = value["id"]
            if value["id"] is None:
                page_layout = html.H1("Welcome!")
            elif value["id"] == "problem_setup_step":
                page_layout = problem_setup_page.layout(problem_setup_step)
            else:
                monitoring_step.selected_actor_from_treeview = extract_dict_by_key(problem_setup_step.project_tree, "uid", value["id"], expect_unique=True, return_index=False)["uid"]
                page_layout = monitoring_page.layout(problem_setup_step, monitoring_step)
                actors_states_ids = json.loads(monitoring_step.actors_states_ids_file.read_text())
                footer_buttons.insert(
                    0,
                        dbc.DropdownMenu(
                        id="selected_state_dropdown",
                        label=f"Selected state: {actors_states_ids[monitoring_step.selected_actor_from_treeview][0]}",
                        size="sm",
                        menu_variant="dark",
                        children=[dbc.DropdownMenuItem(state_id) for state_id in actors_states_ids[monitoring_step.selected_actor_from_treeview]],
                        style={"background-color": "rgba(242, 242, 242, 0.6)", "borderColor": "rgba(242, 242, 242, 0.6)", "color": "rgba(0, 0, 0, 1)"},
                    ),
                )
        footer = [
            dbc.ButtonGroup(
                footer_buttons,
                size="md",
                className="me-1",
            ),
            dbc.Collapse(
                id="osl_logs_collapse",
                is_open=False,
            )
        ]
        return (
            page_layout,
            footer,
            focusRequested
        )
    else:
        raise PreventUpdate


@callback(
    Output("navigation_tree", "items"),
    Input("url", "pathname"),
    Input("trigger_treeview_display", "data"),
    prevent_initial_call=True,
)
def display_tree_view(pathname, trigger_treeview_display):
    """Display treeview with all project nodes."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    if problem_setup_step.project_initialized:
        return problem_setup_step.treeview_items
    else:
        raise PreventUpdate


@callback(
    Output("osl_logs_collapse", "children"),
    Output("osl_logs_collapse", "is_open"),
    Input("osl_logs_button", "n_clicks"),
    Input("url", "pathname"),
    State("osl_logs_collapse", "is_open"),
    prevent_initial_call=True,
)
def display_optislang_logs(n_clicks, pathname, is_open):
    """Display optiSLang logs."""
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    problem_setup_step = project.steps.problem_setup_step

    if not n_clicks:
        # Button has never been clicked
        return None, False

    if problem_setup_step.project_locked:
        osl_logs = problem_setup_step.osl_log_file.read_text().split('\n')
        table = LogsTable(osl_logs)
        return table.render(), not is_open

    raise PreventUpdate


@callback(
    Output("open_in_browser", "children"),
    Input("open_in_browser", "n_clicks"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def open_in_browser(n_clicks, pathname):
    """Open the Portal UI in browser view."""
    portal_ui_url = DashClient.get_portal_ui_url()
    webbrowser.open_new(portal_ui_url)

    raise PreventUpdate



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
PyAnsys Templates.

A collection of interactive templates for building Python projects from scratch
according to PyAnsys guidelines.
"""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__.replace(".", "-"))

AVAILABLE_TEMPLATES_AND_DESCRIPTION = {
    "doc-project": "Create a documentation project using Sphinx.",
    "pybasic": "Create a basic Python Package.",
    "pyansys": "Create a PyAnsys Python Package project.",
    "pyansys-advanced": "Create an advanced PyAnsys Python Package project.",
    "pyansys-openapi_client": "Create an OpenAPI Client Package project.",
    "pyace": "Create a Python project for any method developers.",
    "pyace-flask": "Create a Flask project initialized for any developer.",
    "pyace-grpc": "Create gRPC project initialized for any developer.",
    "pyace-fast": "Create a FastAPI project initialized for any developer.",
    "solution": "[Ansys Internal Use Only] Create a solution based on SAF.",
}
"""A list holding all available templates names."""



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the problem setup step."""

import json
from pathlib import Path
import time
from typing import List, Union, Optional

from ansys.optislang.core import Optislang, logging
from ansys.saf.glow.solution import FileReference, StepModel, StepSpec, instance, long_running, transaction
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.solution.optislang_manager import OptislangManager
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.utilities.common_functions import read_log_file


class MonitoringStep(StepModel):
    """Step model of the monitoring step."""

    # Parameters ------------------------------------------------------------------------------------------------------

    # Frontend persistence
    selected_actor_from_treeview: Optional[str] = None # uid of the actor selected from the treeview.
    selected_command: Optional[str] = None
    selected_actor_from_command: Optional[str] = None
    selected_actor_from_treeview_states_ids: list = []
    commands_locked: bool = False
    auto_update_activated: bool = True
    actor_uid: Optional[str] = None
    project_command_execution_status: dict = {"alert-message": "", "alert-color": "info"}
    actor_command_execution_status: dict = {"alert-message": "", "alert-color": "info"}
    project_btn_group_options: List = [
                                {
                                    "icon": "fas fa-play",
                                    "tooltip": "Restart optiSLang project.",
                                    "value": "restart",
                                    "id": {
                                        "type":"action-button",
                                        "action":"restart"
                                    }
                                },
                                {
                                    "icon":"fa fa-hand-paper",
                                    "tooltip": "Stop optiSLang project gently.",
                                    "value": "stop_gently",
                                    "id": {
                                        "type":"action-button",
                                        "action":"stop_gently"
                                    }
                                },
                                {
                                    "icon": "fas fa-stop",
                                    "tooltip":"Stop optiSLang project.",
                                    "value": "stop",
                                    "id": {
                                        "type": "action-button",
                                        "action":"stop"
                                        }
                                    },
                                {
                                    "icon": "fas fa-fast-backward",
                                    "tooltip": "Reset optiSLang project.",
                                    "value": "reset",
                                    "id": {
                                        "type":"action-button",
                                        "action":"reset"
                                        }
                                    },
                                {
                                    "icon": "fas fa-power-off",
                                    "tooltip": "Shutdown optiSLang project.",
                                    "value": "shutdown",
                                    "id": {
                                        "type":"action-button",
                                        "action":"shutdown"
                                    }
                                },
                            ]
    actor_btn_group_options: List = [
                            {
                                "icon": "fas fa-play",
                                "tooltip": "Restart node.",
                                "value": "restart",
                                "id": {
                                    "type":"action-button",
                                    "action":"restart"
                                }
                            },

                            {
                                "icon":"fa fa-hand-paper",
                                "tooltip": "Stop node gently.",
                                "value": "stop_gently",
                                "id": {
                                    "type":"action-button",
                                    "action":"stop_gently"
                                }
                            },

                            {
                                "icon": "fas fa-stop",
                                "tooltip":"Stop node.",
                                "value": "stop",
                                "id": {
                                    "type": "action-button",
                                    "action":"stop"
                                }
                            },

                            {
                                "icon": "fas fa-fast-backward",
                                "tooltip": "Reset node.",
                                "value": "reset",
                                "id": {
                                    "type":"action-button",
                                    "action":"reset"
                                }
                            },
                        ]

    # Backend data model
    results_files: dict = {}
    project_state: str = "NOT STARTED"
    osl_project_states: List = [
        "IDLE",
        "PROCESSING",
        "PAUSED",
        "PAUSE_REQUESTED",
        "STOPPED",
        "STOP_REQUESTED",
        "GENTLY_STOPPED",
        "GENTLE_STOP_REQUESTED",
        "FINISHED"
    ]
    osl_actor_states: List = [
        "Idle",
        "Succeeded",
        "Failed",
        "Running",
        "Aborted",
        "Predecessor failed",
        "Skipped",
        "Incomplete",
        "Processing done",
        "Stopped",
        "Gently stopped"
    ]
    optislang_logs: List = []
    command_timeout: int = 30

    # File storage ----------------------------------------------------------------------------------------------------

    # Output
    optislang_log_file: FileReference = FileReference("Problem_Setup/pyoptislang.log")
    project_status_info_file: FileReference = FileReference("Problem_Setup/project_status_info.json")
    actors_info_file: FileReference = FileReference("Problem_Setup/actors_info.json")
    actors_status_info_file: FileReference = FileReference("Problem_Setup/actors_status_info.json")
    actors_states_ids_file: FileReference = FileReference("Problem_Setup/actors_states_ids.json")

    # Methods ---------------------------------------------------------------------------------------------------------

    @transaction(
        problem_setup_step=StepSpec(
            download=[
                "osl_server_host",
                "osl_server_port",
                "project_tree",
                "optislang_log_level",
            ]
        ),
        self=StepSpec(
            upload=[
                "project_state",
                "project_status_info_file",
                "actors_info_file",
                "actors_status_info_file",
                "actors_states_ids_file",
                "optislang_logs",
                "optislang_log_file",
                "results_files",
            ],
        )
    )
    @instance("problem_setup_step.osl", identifier="osl")
    @long_running
    def upload_project_data(self, problem_setup_step: ProblemSetupStep, osl: OptislangManager) -> None:
        """Monitor the progress of the optiSLang project and continuously upload project data."""
        # Creat monitoring directory
        Path(self.optislang_log_file.path).parent.mkdir(parents=True, exist_ok=True)

        # Configure logging.
        osl_logger = logging.OslLogger(
            loglevel=problem_setup_step.optislang_log_level,
            log_to_file=True,
            logfile_name=self.optislang_log_file.path,
            log_to_stdout=True,
        )
        osl.instance.__logger = osl_logger.add_instance_logger(osl.instance.name, osl.instance, problem_setup_step.optislang_log_level)

        # Monitor project state and upload data.
        while True:
            # Get project state
            self.project_state = osl.instance.project.get_status()
            osl.instance.log.info(f"Analysis status: {self.project_state}")
            # Get project status info
            with open(self.project_status_info_file.path, "w") as json_file: json.dump(osl.instance.get_osl_server().get_full_project_status_info(), json_file)
            # Read pyoptislang logs
            self.optislang_logs = read_log_file(self.optislang_log_file.path)
            # Get actor status info
            actors_info, actors_status_info, actors_states_ids = {}, {}, {}
            for node_info in problem_setup_step.project_tree:
                if node_info["uid"] == osl.instance.project.root_system.uid:
                    node = osl.instance.project.root_system
                else:
                    node = osl.instance.project.root_system.find_node_by_uid(node_info["uid"], search_depth=-1)
                actors_info[node.uid] = osl.instance.get_osl_server().get_actor_info(node.uid)
                actors_states_ids[node.uid] = node.get_states_ids()
                if node.get_states_ids():
                    actors_status_info[node.uid] = []
                    for hid in node.get_states_ids():
                        actors_status_info[node.uid].append(
                            osl.instance.get_osl_server().get_actor_status_info(node.uid, hid)
                        )
            with open(self.actors_info_file.path, "w") as json_file: json.dump(actors_info, json_file)
            with open(self.actors_status_info_file.path, "w") as json_file: json.dump(actors_status_info, json_file)
            with open(self.actors_states_ids_file.path, "w") as json_file: json.dump(actors_states_ids, json_file)
            # Upload fields
            self.transaction.upload(["project_state"])
            self.transaction.upload(["project_status_info_file"])
            self.transaction.upload(["actors_info_file"])
            self.transaction.upload(["actors_status_info_file"])
            self.transaction.upload(["optislang_logs"])
            self.transaction.upload(["actors_states_ids_file"])
            if self.project_state == "FINISHED":
                break
            time.sleep(3) # Waiting 3 sec before pulling new data. The frequency might be adjusted in the future.


    @transaction(
        problem_setup_step=StepSpec(
            download=[
                "osl_server_host",
                "osl_server_port",
                "project_tree",
            ]
        ),
        self=StepSpec(
            download=[
                "command_timeout",
                "selected_actor_from_command",
                "selected_command",
            ],
            upload=["actor_uid"],
        )
    )
    @instance("problem_setup_step.osl_manager", identifier="osl_manager")
    def control_node_state(self, osl_manager: OptislangManager) -> None:
        """Update the state of root or actor node based on the selected command in the UI."""
        osl = osl_manager.instance
        if not self.selected_actor_from_command == "shutdown":
            if self.selected_actor_from_command == osl.project.root_system.uid:
                node = osl.project.root_system
                self.actor_uid = None
            else:
                node = osl.project.root_system.find_node_by_uid(self.selected_actor_from_command, search_depth=-1)
                self.actor_uid = node
        else:
            node = osl.project.root_system

        status = node.control(self.selected_command, wait_for_completion=True, timeout=self.command_timeout)

        if not status:
            raise Exception(f"{self.selected_command.replace('_', ' ').title()} command against node {node.get_name()} failed.")



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Backend of the problem setup step."""

import json
import os
import platform
import tempfile
import time

from pathlib import Path
from typing import List, Optional, Union

from ansys.optislang.core import Optislang, utils, logging
from ansys.optislang.core.errors import OslCommunicationError
from ansys.optislang.core.osl_server import OslServer
from ansys.saf.glow.solution import (
    FileGroupReference,
    FileReference,
    AssetFileReference,
    StepModel,
    StepSpec,
    create_instance,
    instance,
    long_running,
    transaction,
)
from ansys.solutions.optislang.frontend_components.project_properties import (
    ProjectProperties,
    apply_placeholders_to_properties_file,
    write_properties_file,
)
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.datamodel import datamodel
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.solution.optislang_manager import OptislangManager
from ansys.solutions.{{cookiecutter.__solution_name_slug}}.utilities.common_functions import (
    get_treeview_items_from_project_tree, check_optislang_server, get_states_ids_from_states
)


class ProblemSetupStep(StepModel):
    """Step model of the problem setup step."""

    # Parameters ------------------------------------------------------------------------------------------------------

    # Frontend persistence
    ansys_ecosystem_ready: bool = False
    project_initialized: bool = False
    analysis_locked: bool = False
    project_locked: bool = False
    analysis_started: bool = False

    # Backend data model
    ansys_ecosystem: dict = {
        "optislang": {
            "authorized_versions": [],
            "installed_versions": [],
            "compatible_versions": [],
            "selected_version": None,
            "alert_message": "optiSLang install not checked.",
            "alert_color": "warning",
            "alias": "optiSLang",
        }
    }
    osl_server_host: Optional[str] = None
    osl_server_port: Optional[int] = None
    osl_loglevel: str = "INFO"
    osl_project_tree: list = []
    osl_start_timeout: int = 100 # second
    osl_instance_started: bool = False
    osl_server_healthy: Optional[bool] = None
    osl_project_state: str = "NOT STARTED"
    osl_max_server_request_attempts: int = 3
    placeholders: dict = {}
    registered_files: List = []
    settings: dict = {}
    parameter_manager: dict = {}
    criteria: dict = {}
    ui_placeholders: dict = {}
    app_metadata: dict = {}
    treeview_items: list = [
        {
            "id": "problem_setup_step",
            "text": "Problem Setup",
            "expanded": True,
            "prefixIcon": {
                "src": "https://s2.svgbox.net/hero-solid.svg?ic=adjustments"
            },
            "level": 0
        },
    ]

    # File storage ----------------------------------------------------------------------------------------------------

    # Inputs
    project_file: FileReference = FileReference("Problem_Setup/{{ cookiecutter.__optiSLang_application_archive_stem }}.opf")
    properties_file: AssetFileReference = AssetFileReference(relative_path="Problem_Setup/{{ cookiecutter.__optiSLang_application_archive_stem }}.json", encrypted=False)
    metadata_file: AssetFileReference = AssetFileReference(relative_path="Problem_Setup/metadata.json", encrypted=False)
    input_files: FileGroupReference = FileGroupReference("Problem_Setup/Input_Files/*.*")
    # If folder doesn't exist, it will be created later
    upload_directory: str = os.path.join(tempfile.gettempdir(), "GLOW")

    # Outputs
    working_properties_file: FileReference = FileReference("Problem_Setup/working_properties_file.json")
    osl_log_file: FileReference = FileReference("Problem_Setup/optiSLang.log")
    full_project_status_info_file: FileReference = FileReference("Monitoring/full_project_status_info.json")
    project_data_file: FileReference = FileReference("Monitoring/project_data.json")

    # Methods ---------------------------------------------------------------------------------------------------------

    @transaction(
        self=StepSpec(
            upload=[
                "project_file",
            ]
        )
    )
    def upload_bulk_files_to_project_directory(self) -> None:
        """Upload bulk files to project directory."""

        original_project_file = Path(__file__).absolute().parent.parent / "logic" / "assets" / "{{ cookiecutter.__optiSLang_application_archive_stem }}.opf"
        self.project_file.write_bytes(original_project_file.read_bytes())

    @transaction(
        self=StepSpec(
            download=["metadata_file"],
            upload=["app_metadata"]
        )
    )
    def get_app_metadata(self) -> None:
        """Read OWA metadata file."""
        with open(self.metadata_file.path) as f:
            self.app_metadata = json.load(f)

    @transaction(
        self=StepSpec(
            download=["properties_file"],
            upload=["placeholders", "registered_files", "settings", "parameter_manager", "criteria"]
        )
    )
    def get_default_placeholder_values(self):
        """Get placeholder values and definitions using the ProjectProperties class."""
        pp = ProjectProperties()
        pp.read_file(self.properties_file.path)
        self.placeholders = pp._placeholders
        self.registered_files = pp._registered_files
        self.settings = pp._settings
        self.parameter_manager = pp._parameter_manager
        self.criteria = pp._criteria

    @transaction(
        self=StepSpec(
            download=["properties_file", "ui_placeholders"],
            upload=[
                "working_properties_file",
                "placeholders",
                "registered_files",
                "settings",
                "parameter_manager",
                "criteria",
            ],
        )
    )
    def write_updated_properties_file(self) -> None:
        """Write updated optiSLang project properties file."""
        Path(self.working_properties_file.path).parent.mkdir(parents=True, exist_ok=True)
        properties = apply_placeholders_to_properties_file(self.ui_placeholders, self.properties_file.path)
        self.placeholders = properties["placeholders"]
        self.registered_files = properties["registered_files"]
        self.settings = properties["settings"]
        self.parameter_manager = properties["parameter_manager"]
        self.criteria = properties["criteria"]
        write_properties_file(properties, Path(self.working_properties_file.path))

    @transaction(
        self=StepSpec(
            download=["properties_file", "ui_placeholders"],
            upload=["placeholders", "registered_files", "settings", "parameter_manager", "criteria"],
        )
    )
    def update_osl_placeholders_with_ui_values(self) -> None:
        """Update placeholders with values selected by the user in the UI."""
        properties = apply_placeholders_to_properties_file(self.ui_placeholders, self.properties_file.path)
        self.placeholders = properties["placeholders"]
        self.registered_files = properties["registered_files"]
        self.settings = properties["settings"]
        self.parameter_manager = properties["parameter_manager"]
        self.criteria = properties["criteria"]

    @transaction(
        self=StepSpec(
            upload=["ansys_ecosystem", "ansys_ecosystem_ready"],
        )
    )
    def check_ansys_ecosystem(self) -> None:
        """Check if Ansys Products are installed and if the appropriate versions are available."""
        self.ansys_ecosystem_ready = True

        # Collect optiSLang installations
        self.ansys_ecosystem["optislang"]["installed_versions"] = sorted(list(dict(utils.find_all_osl_exec()).keys()))
        if len(self.ansys_ecosystem["optislang"]["authorized_versions"]) > 0:
            self.ansys_ecosystem["optislang"]["compatible_versions"] = [
                product_version
                for product_version in self.ansys_ecosystem["optislang"]["installed_versions"]
                if product_version in self.ansys_ecosystem["optislang"]["authorized_versions"]
            ]
        else:
            self.ansys_ecosystem["optislang"]["compatible_versions"] = self.ansys_ecosystem["optislang"]["installed_versions"]

        # Check ecosystem
        for product_name in self.ansys_ecosystem.keys():
            if len(self.ansys_ecosystem[product_name]["installed_versions"]) == 0:
                alert_message = f"No installation of {product_name.title()} found in the machine {platform.node()}."
                alert_color = "danger"
                self.ansys_ecosystem_ready = False
            elif len(self.ansys_ecosystem[product_name]["compatible_versions"]) == 0:
                alert_message = (
                    f"None of the authorized versions of {product_name.title()} "
                    f"is installed in the machine {platform.node()}.\n"
                )
                alert_message += "At least one of these versions is required:"
                for authorized_version in self.ansys_ecosystem[product_name]["authorized_versions"]:
                    self.ansys_ecosystem[product_name][
                        "alert_message"
                    ] += f" {authorized_version}"
                alert_message += "."
                alert_color = "danger"
                self.ansys_ecosystem_ready = False
            else:
                self.ansys_ecosystem[product_name]["selected_version"] = self.ansys_ecosystem[product_name][
                    "compatible_versions"
                ][
                    -1
                ]  # Latest
                alert_message = f"{product_name.title()} install detected. Compatible versions are:"
                for compatible_version in self.ansys_ecosystem[product_name]["compatible_versions"]:
                    alert_message += f" {compatible_version}"
                alert_message += ".\n"
                alert_message += "Selected version is %s." % (self.ansys_ecosystem[product_name]["selected_version"])
                alert_color = "success"
            self.ansys_ecosystem[product_name]["alert_message"] = alert_message
            self.ansys_ecosystem[product_name]["alert_color"] = alert_color

    @transaction(
        self=StepSpec(
            download=[
                "project_file",
                "ansys_ecosystem",
                "osl_loglevel",
            ],
            upload=[
                "osl_project_tree",
                "treeview_items"
            ],
        )
    )
    def get_project_tree(self) -> None:
        """Get the project tree of the optiSLang project."""
        # Start optiSLang instance.
        osl = Optislang(
            project_path=self.project_file.path,
            executable=utils.get_osl_exec(self.ansys_ecosystem["optislang"]["selected_version"])[1],
            reset=True,
            shutdown_on_finished=True,
            loglevel=self.osl_loglevel,
            ini_timeout=300,
        )

        # Get project tree
        self.osl_project_tree = osl.project._get_project_tree()

        # Update treeview items
        self.treeview_items = get_treeview_items_from_project_tree(self.osl_project_tree)

    @transaction(
        self=StepSpec(
            download=[
                "input_files",
                "project_file",
                "working_properties_file",
                "osl_loglevel",
                "ansys_ecosystem",
                "osl_project_tree",
                "osl_max_server_request_attempts"
            ],
            upload=[
                "osl_instance_started",
                "osl_server_healthy",
                "osl_server_host",
                "osl_server_port",
                "osl_project_state",
                "osl_log_file",
                "analysis_started",
                "full_project_status_info_file",
                "project_data_file",
            ],
        )
    )
    @long_running
    @create_instance("osl_manager", OptislangManager)
    def start_and_monitor_osl_project(self, osl_manager: OptislangManager) -> None:
        """Start optiSLang and run the project."""
        # Create monitoring directory
        Path(self.project_data_file.path).parent.mkdir(parents=True, exist_ok=True)
        # Start optiSLang instance using instance manager.
        try:
            osl_manager.initialize(
                project_path=self.project_file.path,
                project_properties_file=self.working_properties_file.path,
                osl_version=self.ansys_ecosystem["optislang"]["selected_version"],
                loglevel=self.osl_loglevel
            )
            self.osl_instance_started = True
        except ConnectionRefusedError:
            self.osl_instance_started = False
            return
        self.transaction.upload(["osl_instance_started"])
        # Get optiSLang instance
        osl = osl_manager.instance
        # Set timeout
        osl.set_timeout(25)
        # Get optiSLang server
        osl_server =  osl.get_osl_server()
        # Get server host
        self.osl_server_host = osl_server.get_host()
        # Get server port
        server_info = osl_server.get_server_info()
        self.osl_server_port = server_info["server"]["server_port"]
        # Configure logging.
        osl_logger = logging.OslLogger(
            loglevel=self.osl_loglevel,
            log_to_file=True,
            logfile_name=self.osl_log_file.path,
            log_to_stdout=True,
        )
        osl.__logger = osl_logger.add_instance_logger(osl.name, osl, self.osl_loglevel)
        # Start optiSLang project
        osl.log.info("Start analysis")
        osl.start(wait_for_started=True, wait_for_finished=False)
        # Initialize project data structure
        project_data = {"project": {"information": {}}, "actors": {}}
        for node_props in self.osl_project_tree:
            # Get node
            if node_props["uid"] == osl.project.root_system.uid:
                node = osl.project.root_system
            else:
                node = osl.project.root_system.find_node_by_uid(node_props["uid"], search_depth=-1)
            # Initialize dictionary
            project_data["actors"][node.uid] = {"states_ids": {}, "information": {}, "log": {}, "statistics": {}, "design_table": {}}
        # Monitor project state and upload data.
        while self.osl_project_state not in ["FINISHED", "ABORTED"]:
            # Get project state
            self.osl_project_state = osl.project.get_status()
            osl.log.info(f"Project state: {self.osl_project_state}")
            # Get full project status info
            full_project_status_info = osl_server.get_full_project_status_info()
            with open(self.full_project_status_info_file.path, "w") as json_file: json.dump(full_project_status_info, json_file)
            # Collect project information
            project_data["project"]["information"] = datamodel.extract_project_status_info(full_project_status_info)
            # Walk through project tree
            for node_props in self.osl_project_tree:
                # Get node
                if node_props["uid"] == osl.project.root_system.uid:
                    node = osl.project.root_system
                else:
                    node = osl.project.root_system.find_node_by_uid(node_props["uid"], search_depth=-1)
                # Get actor info (TCP REQUEST)
                actor_info = self._make_osl_server_request(osl_server, "get_actor_info", node.uid)
                # Collect actor log data
                project_data["actors"][node.uid]["log"] = datamodel.extract_actor_log_data(actor_info)
                # Collect actor statistics data
                project_data["actors"][node.uid]["statistics"] = datamodel.extract_actor_statistics_data(actor_info)
                # Get states ids (TCP REQUEST)
                actor_states = self._make_osl_server_request(osl_server, "get_actor_states", node.uid)
                actor_states_ids = get_states_ids_from_states(actor_states)
                project_data["actors"][node.uid]["states_ids"] = actor_states_ids
                # Walk through states ids
                if len(actor_states_ids):
                    for hid in actor_states_ids:
                        # Get actor status info (TCP REQUEST)
                        actor_status_info = self._make_osl_server_request(osl_server, "get_actor_status_info", node.uid, hid)
                        # Collect actor information data
                        project_data["actors"][node.uid]["information"][hid] = datamodel.extract_actor_information_data(actor_status_info, node_props["kind"])
                        # Collect design table data
                        if node_props["kind"] == "system":
                            project_data["actors"][node.uid]["design_table"][hid] = datamodel.extract_design_table_data(actor_status_info)
            # Dump project data
            with open(self.project_data_file.path, "w") as json_file: json.dump(project_data, json_file, allow_nan=True)
            # Upload fields
            self.transaction.upload(["osl_project_state"])
            self.transaction.upload(["full_project_status_info_file"])
            self.transaction.upload(["project_data_file"])
            self.transaction.upload(["osl_log_file"])
            # Wait
            time.sleep(5)

    @transaction(
        self=StepSpec(
            upload=["osl_server_healthy"],
        )
    )
    @instance("osl_manager", identifier="osl_manager")
    def check_optislang_server(self, osl_manager: OptislangManager) -> None:
        """optiSLang server health check."""
        osl = osl_manager.instance
        self.osl_server_healthy = check_optislang_server(osl.get_osl_server())

    def _make_osl_server_request(self, osl_server: OslServer, request_name: str, actor_uid: str, hid: str = None) -> Union[list, dict]:
        """Make a server request and try multiple times if a communication error is raised."""
        # Check inputs
        if request_name not in ["get_actor_info", "get_actor_states", "get_actor_status_info"]:
            raise ValueError(f"Unknown request {request_name}.")
        if request_name == "get_actor_status_info" and not hid:
            raise Exception("A state id is needed to run the get_actor_status_info request.")
        # Get method from optiSLanf server
        request = getattr(osl_server, request_name)
        # Make request
        attempts = 0
        success = False
        while attempts <= self.osl_max_server_request_attempts:
            attempts += 1
            try:
                if request_name == "get_actor_info":
                    response = request(actor_uid)
                elif request_name == "get_actor_states":
                    response = request(actor_uid)
                elif request_name == "get_actor_status_info":
                    response = request(actor_uid, hid)
                success = True
            except OslCommunicationError:
                pass
            if success:
                break
        if not success:
            request_name_splitted = request_name.replace("_", " ").capitalize()
            raise Exception(f"{request_name_splitted} method failed after {attempts} attempts.")

        return response



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
import json
import pandas as pd
import uuid

from dash_extensions.enrich import html, dash_table

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import convert_microseconds


class ActorStatisticsTableAIO(html.Div):

    class ids:
        datatable = lambda aio_id: {
            'component': 'ActorStatisticsTableAIO',
            'subcomponent': 'datatable',
            'aio_id': aio_id
        }

    ids = ids

    def __init__(self, monitoring_step: MonitoringStep, aio_id: str = None):
        """ActorStatisticsTableAIO is an All-in-One component that is composed
        of a parent `html.Div` with a `dcc.Interval` and a `dash_table.DataTable` as children.

        - `monitoring_step` - The StepModel object of the monitoring step.
        - `datatable_props` - A dictionary of properties passed into the dash_table.DataTable component.
        - `interval_props` - A dictionary of properties passed into the dcc.Interval component.
        - `aio_id` - The All-in-One component ID used to generate the table components's dictionary IDs.
        """

        if aio_id is None:
            aio_id = str(uuid.uuid4())

        data = self.get_data(monitoring_step)

        datatable_props = {
            "data": data.to_dict('records'),
            "columns": [{"name": i, "id": i, "type": "text"} for i in data.columns],
            "fixed_rows": {"headers": True},
            "style_header": {
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": "15px",
                "fontWeight": "bold",
            },
            "style_cell": {
                "textAlign": "left",
                "font_family": "Roboto",
                "font_size": "15px",
            },
            "style_cell_conditional": [
                {"if": {"column_id": "row_names"}, "minWidth": "50px", "maxWidth": "50px", "width": "50px"},
                {"if": {"column_id": "Current Run"}, "minWidth": "50px", "maxWidth": "50px", "width": "50px"},
                {"if": {"column_id": "All Runs"}, "minWidth": "50px", "maxWidth": "50px", "width": "50px"},
            ],
            "style_data_conditional": [
                {
                    "if": {"column_id": "row_names", "filter_query": '{Level} eq "std_dev"'},
                    "backgroundColor": "rgb(227, 245, 252)",
                    "color": "rgb(0, 0, 0)",
                    "textAlign": "center",
                }
            ],
            "style_as_list_view": True,
        }

        super().__init__([
            dbc.Card(
                [
                    dbc.CardBody(
                        [
                            html.H4("Actor Statistics", className="card-title"),
                            dash_table.DataTable(id=self.ids.datatable(aio_id), **datatable_props),
                        ]
                    ),
                ],
                color="warning",
                outline=True,
            )
        ])

    def get_data(self, monitoring_step: MonitoringStep) -> pd.DataFrame:

        actors_info = json.loads(monitoring_step.actors_info_file.read_text())

        if monitoring_step.selected_actor_from_treeview in actors_info.keys():
            actor_info = actors_info[monitoring_step.selected_actor_from_treeview]
        else:
            actor_info = {}

        actor_statistics_data = {
            "row_names": ["Usages", "Accumulated", "Minimum", "Maximum", "Mean", "std_dev"],
            "Current Run": [None, None, None, None, None, None],
            "All Runs": [None, None, None, None, None, None],
        }

        if actor_info:
            for column_name in ["Current Run", "All Runs"]:
                key = column_name.split()[0].lower()
                actor_statistics_data[column_name][0] = actor_info["usage_stats"][key]["num_usages"]
                for index, row_name in enumerate(actor_statistics_data["row_names"]):
                    if row_name != "Usages":
                        duration = actor_info["usage_stats"][key]["exec_duration_us"][row_name.lower()]
                        duration = convert_microseconds(duration)
                        actor_statistics_data[column_name][index] = duration

        return pd.DataFrame(actor_statistics_data)



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Solution definition module."""


from ansys.saf.glow.solution import Solution, StepsModel

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.first_step import FirstStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.intro_step import IntroStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.second_step import SecondStep


class Steps(StepsModel):
    """Workflow definition."""
    intro_step: IntroStep
    first_step: FirstStep
    second_step: SecondStep


class {{ cookiecutter.__solution_definition_name }}(Solution):
    """Solution definition."""
    display_name: str = "{{ cookiecutter.solution_display_name.replace('"', '') }}"
    steps: Steps



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import sys

from typing import List

from ansys.saf.glow.desktop.orchestration.pim_configurations import PimProductConfigurationBuilder
from ansys.saf.glow.desktop.orchestration.pim_process import PimProductConfiguration


class OptislangConfigBuilder(PimProductConfigurationBuilder):
    def get_configurations(self) -> List[PimProductConfiguration]:
        return [
            PimProductConfiguration(
                "optislang",
                str(sys.executable),
                arguments=["-m", "uvicorn", "ansys.solutions.{{ cookiecutter.__solution_name_slug }}.pim_configurations.osl_wrapper:app", "--port", "${AENEID_PORT_HTTP}"],
                service_name="http",
                service_type="http",
            )
        ]



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the design table view."""

import dash_bootstrap_components as dbc
from dash_extensions.enrich import html, Input, Output, State, dcc

from ansys.saf.glow.client.dashclient import DashClient, callback

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.design_table import DesignTableAIO


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the design table view."""

    return html.Div(
        [
            html.Br(),
            dbc.Row(
                [
                    dbc.Col(html.Div(DesignTableAIO(monitoring_step), id="design_table"), width=12),
                ]
            ),
            dcc.Interval(
                id="design_table_auto_update",
                interval=monitoring_step.auto_update_frequency,  # in milliseconds
                n_intervals=0,
                disabled=False if monitoring_step.auto_update_activated else True
            ),
        ]
    )


@callback(
    Output("design_table_auto_update", "disabled"),
    Input("activate_auto_update", "on"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def activate_auto_update(on, pathname):
    """Enable/Disable auto update."""
    return not on


@callback(
    Output("design_table", "children"),
    Input("design_table_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_view(n_intervals, pathname):
    """Update design table."""
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    # Get problem setup step
    problem_setup_step = project.steps.problem_setup_step
    # Get monitoring step
    monitoring_step = project.steps.monitoring_step

    return DesignTableAIO(monitoring_step)



{%- if cookiecutter.copyright != "None" -%}
# Copyright (c) {% now "utc", '%Y' %}, {{ cookiecutter.copyright }}. Unauthorised use, distribution or duplication is prohibited
{% endif %}
"""
{{ cookiecutter.project_name }}.

{{ cookiecutter.library_name }}
"""
from datetime import datetime
import logging
import os


class Logger(object):
    """Logger class."""

    @classmethod
    def init(cls, name):
        """Initialize logger instances."""
        formatter = logging.Formatter(
            "%(asctime)s %(name)s %(levelname)s %(threadName)-10s %(message)s"
        )

        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)

        filename = "{0}.log".format(datetime.today().strftime("%Y_%m_%d@%H_%M_%S"))

        if os.name != "nt":
            filepath = f"/var/log/{filename}"
        else:
            folder = os.path.join(
                os.getenv("LOCALAPPDATA"),
                "{{ cookiecutter.project_name }}",
                "{{ cookiecutter.library_name }}",
            )
            if not os.path.exists(folder):
                os.makedirs(folder)
            filepath = os.path.join(folder, filename)
        fh = logging.FileHandler(filepath)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        return logger



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Solution definition module."""

from ansys.saf.glow.solution import Solution, StepsModel

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep


class Steps(StepsModel):
    """Workflow definition."""
    problem_setup_step: ProblemSetupStep
    monitoring_step: MonitoringStep


class {{ cookiecutter.__solution_definition_name }}(Solution):
    """Solution definition."""
    display_name : str = "{{ cookiecutter.__solution_display_name.replace('"', '') }}"
    steps: Steps



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

from datetime import datetime

from ansys.solutions.{{cookiecutter.__solution_name_slug}}.utilities.common_functions import (
    remove_key_from_dictionaries,
    sort_dict_by_ordered_keys,
    convert_microseconds
)


def extract_project_status_info(full_project_status_info: dict) -> dict:
    """ Extract project status info from the optiSLang server response.

    Parameters
    ----------
    full_project_status_info: dict
        Full project status info.

    Returns
    -------
        dict
            A dictionary containing project status data.
    """

    data = {
        "State": None,
        "Id": None,
        "Name": None,
        "Machine": None,
        "Location": None,
        "Project directory": None,
        "Owner": None,
        "Registered": "--",
        "Lock info": "--"
    }

    if full_project_status_info:
        data["State"] = full_project_status_info["projects"][0].get("state", None)
        data["Id"] = full_project_status_info["projects"][0].get("project_id", None)
        data["Name"] = full_project_status_info["projects"][0].get("name", None)
        data["Machine"] = full_project_status_info["projects"][0].get("machine", None)
        data["Location"] = full_project_status_info["projects"][0].get("location", None)
        data["Project directory"] = full_project_status_info["projects"][0].get("working_dir", None)
        data["Owner"] = full_project_status_info["projects"][0].get("user", None)
        data["Registered"] = "--"
        data["Lock info"] = "--"

    return data


def extract_actor_information_data(actor_status_info: dict, kind: str) -> dict:
    """ Extract actor information data from the optiSLang server response.

    Parameters
    ----------
    actor_status_info: dict
        Actor status info.
    kind: str
        Actor kind (actor/system).

    Returns
    -------
        dict
            A dictionary containing actor information data.
    """

    data = {
        "Working directory": None,
        "Processing state": None,
        "Execution duration": None,
    }
    if kind == "system":
        data["Processed"] = None
        data["Status"] = None
        data["Succeeded"] = None
        data["Not succeeded"] = None
        data["Pending"] = None

    if actor_status_info:
        data["Working directory"] = actor_status_info.get("working dir", None)
        data["Processing state"] = actor_status_info.get("state", None)
        data["Execution duration"] = actor_status_info.get("exec_dur_us", None)
        if kind == "system":
            # Process data
            succeeded_designs = actor_status_info["succeeded_designs"]
            failed_designs = actor_status_info["failed_designs"]
            pending_designs = actor_status_info["pending_designs"]
            total_designs = actor_status_info["total_designs"]
            processed_designs = total_designs - pending_designs
            try:
                status = int(processed_designs / total_designs * 100)
            except ZeroDivisionError:
                status = 0
            # Update data
            data["Processed"] = f"{processed_designs} / {total_designs}"
            data["Status"] = f"{status}%"
            data["Succeeded"] = succeeded_designs
            data["Not succeeded"] = failed_designs
            data["Pending"] = pending_designs

    return data


def extract_actor_log_data(actor_info: dict) -> dict:
    """ Extract actor log data from the optiSLang server response.

    Parameters
    ----------
    actor_info: dict
        Actor info.

    Returns
    -------
        dict
            A dictionary containing actor log data.
    """

    data = {"Time": [], "Level": [], "Message": []}

    if actor_info:
        if "log_messages" in actor_info.keys():
            if len(actor_info["log_messages"]):
                # Remove hid key from list dictionaries because it is useless for UI
                # and prevent transformation to DataFrame.
                data = remove_key_from_dictionaries(actor_info["log_messages"], "hid")
                # Transform list of dictionaries into dictionary
                data = {
                    key: [d[key] for d in actor_info["log_messages"]]
                    for key in actor_info["log_messages"][0]
                }
                # Sort keys in order
                data = sort_dict_by_ordered_keys(data, ["time_stamp", "level", "message"])
                # Rename keys
                data["Time"] = data.pop("time_stamp")
                data["Level"] = data.pop("level")
                data["Message"] = data.pop("message")
                # Convert timestamp
                for index, value in enumerate(data["Time"]):
                    dt = datetime.strptime(value, "%Y%m%dT%H%M%S.%f")
                    data["Time"][index] = dt.strftime("%Y-%m-%d %H-%M-%S-%f")[:-3]

    return data


def extract_actor_statistics_data(actor_info: dict) -> dict:
    """ Extract actor statistics data from the optiSLang server response.

    Parameters
    ----------
    actor_info: dict
        Actor info.

    Returns
    -------
        dict
            A dictionary containing actor statistics data.
    """

    data = {
        "Header": ["Current Run", "All Runs"],
        "Usages": [None, None],
        "Accumulated": [None, None],
        "Minimum": [None, None],
        "Maximum": [None, None],
        "Mean": [None, None],
        "std_dev": [None, None],
    }

    if actor_info:
        for i, run in enumerate(["Current Run", "All Runs"]):
            run_type = run.split()[0].lower()
            data["Usages"][i] = actor_info["usage_stats"][run_type]["num_usages"]
            for key in data.keys():
                if key != "Header" and key != "Usages":
                    duration = actor_info["usage_stats"][run_type]["exec_duration_us"][key.lower()]
                    duration = convert_microseconds(duration)
                    data[key][i] = duration

    data["Standard devieation"] = data.pop("std_dev")

    return data


def extract_design_table_data(actor_status_info: dict) -> dict:
    """Extract design table data from the optiSLang server response.

    Parameters
    ----------
    actor_status_info: dict
        Actor status info.

    Returns
    -------
        dict
            A dictionary containing design table data.
    """

    data = {
        "Design": [],
        "Feasible": [],
        "Status": [],
        "Pareto": [],
    }

    field_types = ["parameter", "constraint", "objective", "response"]

    if actor_status_info:
        for field_type in field_types:
            field_name_index = field_type + "_names"
            for field_name in actor_status_info["designs"][field_name_index]:
                data[field_name] = []
        design_numbers = []
        for design_values in reversed(actor_status_info["designs"]["values"]):
            design_numbers.append(design_values["hid"].split(".")[-1])
            data["Design"].append(design_values["hid"])
            for field_type in field_types:
                field_name_index = field_type + "_names"
                field_value_index = field_type + "_values"
                field_names = actor_status_info["designs"][field_name_index]
                field_values = design_values[field_value_index]
                if len(field_names) != len(field_values):
                    field_values = [None for i in range(len(field_names))]
                for field_name, field_value in zip(field_names, field_values):
                    if isinstance(field_value, dict):
                        if field_value["type"] == "xy_data":
                            field_value = f"[1:%s]" % (field_value["num_entries"])
                    data[field_name].append(field_value)
        for design_status in reversed(actor_status_info["design_status"]):
            data["Feasible"].append(design_status["feasible"])
            data["Status"].append(design_status["status"])
            data["Pareto"].append(design_status["pareto_design"])

    return data



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

import dash_bootstrap_components as dbc
from dash_extensions.enrich import html


class ButtonGroup:
    def __init__(self, options, disabled=False):
        self.options = options
        self.disabled = disabled
        self.buttons = self._create_button_group()

    def _create_button_group(self):
        btn_group = []
        for option in self.options:
            icon = option['icon']
            tooltip = option['tooltip']
            id = option['id']
            button = ButtonWithTooltip(id, icon, self.disabled, tooltip)
            btn_group.append(button)
        return btn_group

class ButtonWithTooltip(html.Div):

    def __init__(self, id, button_icon, is_disabled, tooltip):

        button_style = {
            "display": "flex",
            "justify-content": "center",
            "align-items": "center",
            "fontSize": "150%",
            "color": "rgba(0, 0, 0, 1)",
            "background-color": "rgba(255, 255, 255, 1)",
            "border-color": "rgba(0, 0, 0, 1)",
            "height": "40px",
            "width": "70px",
        }

        super().__init__(
            [
                dbc.Button(
                    id=id,
                    children=html.Div(
                        [
                            html.I(className=button_icon, style={"display": "inline-block"}),
                        ],
                        ),
                    style=button_style,
                    disabled=is_disabled,
                ),
                dbc.Tooltip(
                    id=f"tooltip-{id}",
                    target=id,
                    placement="bottom",
                    children=tooltip
                ),
            ]
        )



"""Project installation script."""

from setuptools import find_namespace_packages, setup

setup(
    name="{{ cookiecutter.__pkg_name }}",
    version="{{ cookiecutter.__version }}",
    url="{{ cookiecutter.__repository_url }}",
    author="ANSYS, Inc.",
    author_email="pyansys.support@ansys.com",
    maintainer="PyAnsys developers",
    maintainer_email="pyansys.core@ansys.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license="MIT",
    license_file="LICENSE",
    description="{{ cookiecutter.__short_description }}",
    long_description=open("README.rst").read(),
    install_requires=["importlib-metadata >=4.0"],
    python_requires=">={{ cookiecutter.__requires_python }}",
    {%- if cookiecutter.__is_pyansys == "True" %}
    packages=find_namespace_packages(where="src", include="ansys*"),
    {%- else %}
    packages=find_namespace_packages(where="src"),
    {%- endif %}
    package_dir={"": "src"},
)



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the summary view."""

import dash_bootstrap_components as dbc

from dash_extensions.enrich import html, Input, Output, State, dcc

from ansys.saf.glow.client.dashclient import DashClient, callback

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.definition import {{ cookiecutter.__solution_definition_name }}
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_information_table import ActorInformationTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.button_group import ButtonGroup
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.node_control import NodeControlAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.system_files import SystemFilesAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_logs_table import ActorLogsTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.ui.components.actor_statistics_table import ActorStatisticsTableAIO
from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.utilities.common_functions import extract_dict_by_key


def layout(problem_setup_step: ProblemSetupStep, monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the summary view."""

    alert_props = {
        "children": monitoring_step.actor_command_execution_status["alert-message"],
        "color": monitoring_step.actor_command_execution_status["alert-color"],
        "is_open": bool(monitoring_step.actor_command_execution_status["alert-message"]),
    }
    button_group = ButtonGroup(options=monitoring_step.actor_btn_group_options, disabled=monitoring_step.commands_locked).buttons
    actor_info = extract_dict_by_key(problem_setup_step.project_tree, "uid", monitoring_step.selected_actor_from_treeview, expect_unique=True, return_index=False)

    content = [
        html.Br(),
        dbc.Row(
            [
                dbc.Col(html.Div(ActorInformationTableAIO(monitoring_step), id="actor_information_table"), width=12),
            ]
        ),
        html.Br(),
        dbc.Row(
            [
                dbc.Col(NodeControlAIO(button_group, alert_props, "actor-commands"), width=12),
            ]
        ),
    ]

    if actor_info:
        if actor_info["kind"] == "system":
            content.extend(
                [
                    html.Br(),
                    dbc.Row(
                        [
                            dbc.Col(SystemFilesAIO(monitoring_step, aio_id="system_files"), width=12),
                        ]
                    ),
                ]
            )

    content.extend(
            [
                html.Br(),
                dbc.Row(
                    [
                        dbc.Col(html.Div(ActorLogsTableAIO(monitoring_step), id="actor_logs_table"), width=12),
                    ]
                ),
                html.Br(),
                dbc.Row(
                    [
                        dbc.Col(html.Div(ActorStatisticsTableAIO(monitoring_step), id="actor_statistics_table"), width=12),
                    ]
                ),
                dcc.Interval(
                    id="summary_auto_update",
                    interval=monitoring_step.auto_update_frequency,  # in milliseconds
                    n_intervals=0,
                    disabled=False if monitoring_step.auto_update_activated else True
                ),
            ]
        )

    return html.Div(content)


@callback(
    Output("summary_auto_update", "disabled"),
    Input("activate_auto_update", "on"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def activate_auto_update(on, pathname):
    """Enable/Disable auto update."""
    return not on


@callback(
    Output("actor_information_table", "children"),
    Output("actor_logs_table", "children"),
    Output("actor_statistics_table", "children"),
    Output("selected_state_dropdown", "options"),
    Output("selected_state_dropdown", "value"),
    Input("summary_auto_update", "n_intervals"),
    State("url", "pathname"),
    prevent_initial_call=True,
)
def update_view(n_intervals, pathname):
    """Update design table."""
    # Get project
    project = DashClient[{{ cookiecutter.__solution_definition_name }}].get_project(pathname)
    monitoring_step = project.steps.monitoring_step

    return (
        ActorInformationTableAIO(monitoring_step),
        ActorLogsTableAIO(monitoring_step),
        ActorStatisticsTableAIO(monitoring_step)
    )



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the scenery view."""

import json
import optislang_dash_lib

from dash_extensions.enrich import html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.problem_setup_step import ProblemSetupStep


def layout(problem_setup_step: ProblemSetupStep) -> html.Div:
    """Layout of the scenery view."""

    project_status_info = json.loads(monitoring_step.project_status_info_file.read_text())

    if project_status_info:
        return html.Div(
            [
                html.Br(),
                html.Div(
                    [
                        optislang_dash_lib.SceneryComponent(
                            id="input",
                            project_state=project_status_info,
                        ),
                    ]
                ),
            ]
        )
    else:
        return html.Div([])



# ©2023, ANSYS Inc. Unauthorized use, distribution or duplication is prohibited.

"""Frontend of the results file view."""

from dash_extensions.enrich import html

from ansys.solutions.{{ cookiecutter.__solution_name_slug }}.solution.monitoring_step import MonitoringStep


def layout(monitoring_step: MonitoringStep) -> html.Div:
    """Layout of the result files view."""

    return html.Div([])
